@@ -3,6 +3,7 @@ import sys
 import string
 import traceback
 import keyword
+import struct
 
 class VarList:
     """Nicely format a C variable list"""
@@ -140,18 +141,6 @@ class CharArg(ArgType):
 	info.varlist.add('gchar', 'ret')
         info.codeafter.append('    return PyString_FromStringAndSize(&ret, 1);')
 class GUniCharArg(ArgType):
-    param_tmpl = ('    if (py_%(name)s[1] != 0) {\n'
-                  '        PyErr_SetString(PyExc_TypeError, "%(name)s should be a 1 character unicode string");\n'
-                  '        return NULL;\n'
-                  '    }\n'
-                  '    %(name)s = (gunichar)py_%(name)s[0];\n')
-    dflt_tmpl = ('    if (py_%(name)s != NULL) {\n'
-                 '        if (py_%(name)s[1] != 0) {\n'
-                 '            PyErr_SetString(PyExc_TypeError, "%(name)s should be a 1 character unicode string");\n'
-                 '            return NULL;\n'
-                 '        }\n'
-                 '        %(name)s = (gunichar)py_%(name)s[0];\n'
-                 '     }\n')
     ret_tmpl = ('#if !defined(Py_UNICODE_SIZE) || Py_UNICODE_SIZE == 2\n'
                 '    if (ret > 0xffff) {\n'
                 '        PyErr_SetString(PyExc_RuntimeError, "returned character can not be represented in 16-bit unicode");\n'
@@ -163,13 +152,10 @@ class GUniCharArg(ArgType):
     def write_param(self, ptype, pname, pdflt, pnull, info):
 	if pdflt:
 	    info.varlist.add('gunichar', pname + " = '" + pdflt + "'")
-            info.codebefore.append(self.dflt_tmpl % {'name':pname})
 	else:
 	    info.varlist.add('gunichar', pname)
-            info.codebefore.append(self.param_tmpl % {'name':pname})
-        info.varlist.add('Py_UNICODE', '*py_' + pname + ' = NULL')
 	info.arglist.append(pname)
-        info.add_parselist('u', ['&py_' + pname], [pname])
+        info.add_parselist('O&', ['pyg_pyobj_to_unichar_conv', '&' + pname], [pname])
     def write_return(self, ptype, ownsreturn, info):
         info.varlist.add('gunichar', 'ret')
         info.varlist.add('Py_UNICODE', 'py_ret')
@@ -188,23 +174,120 @@ class IntArg(ArgType):
         info.varlist.add('int', 'ret')
         info.codeafter.append('    return PyInt_FromLong(ret);')
 
-class ULongArg(ArgType):
-    dflt = '    if (py_%(name)s)\n' \
-           '        %(name)s = PyLong_AsUnsignedLong(py_%(name)s);\n'
-    before = '    %(name)s = PyLong_AsUnsignedLong(py_%(name)s);\n'
+class UIntArg(ArgType):
     def write_param(self, ptype, pname, pdflt, pnull, info):
-        if pdflt:
-            info.varlist.add('gulong', pname + ' = ' + pdflt)
-            info.codebefore.append(self.dflt % {'name':pname})            
-        else:
-            info.varlist.add('gulong', pname)
-            info.codebefore.append(self.before % {'name':pname})            
-        info.varlist.add('PyObject', "*py_" + pname + ' = NULL')
-        info.arglist.append(pname)
-        info.add_parselist('O!', ['&PyLong_Type', '&py_' + pname], [pname])
+	if pdflt:
+	    info.varlist.add(ptype, pname + ' = ' + pdflt)
+	else:
+	    info.varlist.add(ptype, pname)
+	info.arglist.append(pname)
+        info.add_parselist('I', ['&' + pname], [pname])
     def write_return(self, ptype, ownsreturn, info):
-        info.varlist.add('gulong', 'ret')
-        info.codeafter.append('    return PyLong_FromUnsignedLong(ret);')
+        info.varlist.add(ptype, 'ret')
+        info.codeafter.append('    return PyLong_FromUnsignedLong(ret);\n')
+
+class SizeArg(ArgType):
+
+    if struct.calcsize('P') <= struct.calcsize('l'):
+        llp64 = True
+    else:
+        llp64 = False
+    
+    def write_param(self, ptype, pname, pdflt, pnull, info):
+	if pdflt:
+	    info.varlist.add(ptype, pname + ' = ' + pdflt)
+	else:
+	    info.varlist.add(ptype, pname)
+	info.arglist.append(pname)
+        if self.llp64:
+            info.add_parselist('k', ['&' + pname], [pname])
+        else:
+            info.add_parselist('K', ['&' + pname], [pname])
+    def write_return(self, ptype, ownsreturn, info):
+        info.varlist.add(ptype, 'ret')
+        if self.llp64:
+            info.codeafter.append('    return PyLong_FromUnsignedLongLong(ret);\n')
+        else:
+            info.codeafter.append('    return PyLong_FromUnsignedLong(ret);\n')
+
+class SSizeArg(ArgType):
+
+    if struct.calcsize('P') <= struct.calcsize('l'):
+        llp64 = True
+    else:
+        llp64 = False
+    
+    def write_param(self, ptype, pname, pdflt, pnull, info):
+	if pdflt:
+	    info.varlist.add(ptype, pname + ' = ' + pdflt)
+	else:
+	    info.varlist.add(ptype, pname)
+	info.arglist.append(pname)
+        if self.llp64:
+            info.add_parselist('l', ['&' + pname], [pname])
+        else:
+            info.add_parselist('L', ['&' + pname], [pname])
+    def write_return(self, ptype, ownsreturn, info):
+        info.varlist.add(ptype, 'ret')
+        if self.llp64:
+            info.codeafter.append('    return PyLong_FromLongLong(ret);\n')
+        else:
+            info.codeafter.append('    return PyLong_FromLong(ret);\n')
+
+class LongArg(ArgType):
+    def write_param(self, ptype, pname, pdflt, pnull, info):
+	if pdflt:
+	    info.varlist.add(ptype, pname + ' = ' + pdflt)
+	else:
+	    info.varlist.add(ptype, pname)
+	info.arglist.append(pname)
+        info.add_parselist('l', ['&' + pname], [pname])
+    def write_return(self, ptype, ownsreturn, info):
+        info.varlist.add(ptype, 'ret')
+        info.codeafter.append('    return PyInt_FromLong(ret);\n')
+
+class BoolArg(IntArg):
+    def write_return(self, ptype, ownsreturn, info):
+        info.varlist.add('int', 'ret')
+        info.codeafter.append('    return PyBool_FromLong(ret);\n')
+
+class TimeTArg(ArgType):
+    def write_param(self, ptype, pname, pdflt, pnull, info):
+	if pdflt:
+	    info.varlist.add('time_t', pname + ' = ' + pdflt)
+	else:
+	    info.varlist.add('time_t', pname)
+	info.arglist.append(pname)
+        info.add_parselist('i', ['&' + pname], [pname])
+    def write_return(self, ptype, ownsreturn, info):
+        info.varlist.add('time_t', 'ret')
+        info.codeafter.append('    return PyInt_FromLong(ret);')
+
+class ULongArg(ArgType):
+    def write_param(self, ptype, pname, pdflt, pnull, info):
+	if pdflt:
+	    info.varlist.add('unsigned long', pname + ' = ' + pdflt)
+	else:
+	    info.varlist.add('unsigned long', pname)
+	info.arglist.append(pname)
+        info.add_parselist('k', ['&' + pname], [pname])
+    def write_return(self, ptype, ownsreturn, info):
+        info.varlist.add(ptype, 'ret')
+        info.codeafter.append('    return PyLong_FromUnsignedLong(ret);\n')
+
+class UInt32Arg(ULongArg):
+    def write_param(self, ptype, pname, pdflt, pnull, info):
+        ULongArg.write_param(self, ptype, pname, pdflt, pnull, info)
+        ## if sizeof(unsigned long) > sizeof(unsigned int), we need to
+        ## check the value is within guint32 range
+        if struct.calcsize('L') > struct.calcsize('I'):
+            info.codebefore.append((
+                '    if (%(pname)s > G_MAXUINT32) {\n'
+                '        PyErr_SetString(PyExc_ValueError,\n'
+                '                        "Value out of range in conversion of"\n'
+                '                        " %(pname)s parameter to unsigned 32 bit integer");\n'
+                '        return NULL;\n'
+                '    }\n') % vars())
 
 class Int64Arg(ArgType):
     def write_param(self, ptype, pname, pdflt, pnull, info):
@@ -313,7 +396,7 @@ class EnumArg(ArgType):
         info.add_parselist('O', ['&py_' + pname], [pname]);
     def write_return(self, ptype, ownsreturn, info):
         info.varlist.add('gint', 'ret')
-        info.codeafter.append('    return PyInt_FromLong(ret);')
+        info.codeafter.append('    return pyg_enum_from_gtype(%s, ret);' % self.typecode)
 
 class FlagsArg(ArgType):
     flag = ('    if (%(default)spyg_flags_get_value(%(typecode)s, py_%(name)s, (gint *)&%(name)s))\n'
@@ -336,7 +419,7 @@ class FlagsArg(ArgType):
         info.add_parselist('O', ['&py_' + pname], [pname])
     def write_return(self, ptype, ownsreturn, info):
         info.varlist.add('guint', 'ret')
-        info.codeafter.append('    return PyInt_FromLong(ret);')
+        info.codeafter.append('    return pyg_flags_from_gtype(%s, ret);' % self.typecode)
 
 class ObjectArg(ArgType):
     # should change these checks to more typesafe versions that check
@@ -384,7 +467,7 @@ class ObjectArg(ArgType):
 		info.codebefore.append(self.dflt % {'name':pname,
                                                     'cast':self.cast}) 
 		info.arglist.append(pname)
-                info.add_parselist('O', ['&Py%s_Type' % self.objname,
+                info.add_parselist('O!', ['&Py%s_Type' % self.objname,
                                          '&py_' + pname], [pname])
 	    else:
 		info.varlist.add('PyGObject', '*' + pname)
@@ -396,8 +479,9 @@ class ObjectArg(ArgType):
         info.varlist.add(ptype, '*ret')
         if ownsreturn:
             info.varlist.add('PyObject', '*py_ret')
-            info.codeafter.append('    py_ret = pygobject_new((GObject *)ret;\n'
-                                  '    g_object_unref(ret);\n'
+            info.codeafter.append('    py_ret = pygobject_new((GObject *)ret);\n'
+                                  '    if (ret != NULL)\n'
+                                  '        g_object_unref(ret);\n'
                                   '    return py_ret;')
         else:
             info.codeafter.append('    /* pygobject_new handles NULL checking */\n' +
@@ -535,18 +619,28 @@ class PointerArg(ArgType):
                                   '    return pyg_pointer_new(' + self.typecode + ', &ret);')
 
 class AtomArg(IntArg):
+    dflt = '    if (py_%(name)s) {\n' \
+           '        %(name)s = pygdk_atom_from_pyobject(py_%(name)s);\n' \
+           '        if (PyErr_Occurred())\n' \
+           '            return NULL;\n' \
+           '    }\n'
     atom = ('    %(name)s = pygdk_atom_from_pyobject(py_%(name)s);\n'
             '    if (PyErr_Occurred())\n'
             '        return NULL;\n')
     def write_param(self, ptype, pname, pdflt, pnull, info):
-        info.varlist.add('GdkAtom', pname)
-	info.varlist.add('PyObject', '*py_' + pname + ' = NULL')
-	info.codebefore.append(self.atom % {'name': pname})
-	info.arglist.append(pname)
+        if pdflt:
+            info.varlist.add('GdkAtom', pname + ' = ' + pdflt)
+            info.varlist.add('PyObject', '*py_' + pname + ' = NULL')
+            info.codebefore.append(self.dflt % {'name': pname})
+        else:
+            info.varlist.add('GdkAtom', pname)
+            info.varlist.add('PyObject', '*py_' + pname + ' = NULL')
+            info.codebefore.append(self.atom % {'name': pname})
+        info.arglist.append(pname)
         info.add_parselist('O', ['&py_' + pname], [pname])
     def write_return(self, ptype, ownsreturn, info):
         info.varlist.add('GdkAtom', 'ret')
-        info.codeafter.append('    return PyGdkAtom_New(ret);')
+        info.codeafter.append('    return PyString_FromString(gdk_atom_name(ret));')
 
 class GTypeArg(ArgType):
     gtype = ('    if ((%(name)s = pyg_type_from_object(py_%(name)s)) == 0)\n'
@@ -584,12 +678,6 @@ class GtkTreePathArg(ArgType):
             '            return NULL;\n'
             '        }\n'
             '    }\n')
-    null = ('    if (PyTuple_Check(py_%(name)s))\n'
-            '        %(name)s = pygtk_tree_path_from_pyobject(py_%(name)s);\n'
-            '    else if (py_%(name)s != Py_None) {\n'
-            '        PyErr_SetString(PyExc_TypeError, "%(name)s should be a GtkTreePath or None");\n'
-            '        return NULL;\n'
-            '    }\n')
     freepath = ('    if (%(name)s)\n'
                 '        gtk_tree_path_free(%(name)s);\n')
     def __init__(self):
@@ -627,26 +715,79 @@ class GtkTreePathArg(ArgType):
                                   '    return Py_None;')
 						       
 class GdkRectanglePtrArg(ArgType):
-    null = ('    if (!pygdk_rectangle_from_pyobject(py_%(name)s, &%(name)s))\n'
-            '        return NULL;\n')
+    normal = ('    if (!pygdk_rectangle_from_pyobject(py_%(name)s, &%(name)s))\n'
+              '        return NULL;\n')
+    null =   ('    if (py_%(name)s == Py_None)\n'
+              '        %(name)s = NULL;\n'
+              '    else if (pygdk_rectangle_from_pyobject(py_%(name)s, &%(name)s_rect))\n'
+              '        %(name)s = &%(name)s_rect;\n'
+              '    else\n'
+              '            return NULL;\n')
     def write_param(self, ptype, pname, pdflt, pnull, info):
-        info.varlist.add('GdkRectangle', pname + ' = { 0, 0, 0, 0 }')
-        info.varlist.add('PyObject', '*py_' + pname + ' = Py_None')
-        info.add_parselist('O', ['&py_' + pname], [pname])
-	info.arglist.append('&' + pname)
-	info.codebefore.append(self.null % {'name':  pname})
+        if pnull:
+            info.varlist.add('GdkRectangle', pname + '_rect = { 0, 0, 0, 0 }')
+            info.varlist.add('GdkRectangle', '*' + pname)
+            info.varlist.add('PyObject', '*py_' + pname + ' = Py_None')
+            info.add_parselist('O', ['&py_' + pname], [pname])
+            info.arglist.append(pname)
+            info.codebefore.append(self.null % {'name':  pname})
+        else:
+            info.varlist.add('GdkRectangle', pname + ' = { 0, 0, 0, 0 }')
+            info.varlist.add('PyObject', '*py_' + pname)
+            info.add_parselist('O', ['&py_' + pname], [pname])
+            info.arglist.append('&' + pname)
+            info.codebefore.append(self.normal % {'name':  pname})
 
 class GdkRectangleArg(ArgType):
     def write_return(self, ptype, ownsreturn, info):
 	info.varlist.add('GdkRectangle', 'ret')
 	info.codeafter.append('    return pyg_boxed_new(GDK_TYPE_RECTANGLE, &ret, TRUE, TRUE);')
 
+class PyObjectArg(ArgType):
+    def write_param(self, ptype, pname, pdflt, pnull, info):
+        info.varlist.add('PyObject', '*' + pname)
+        info.add_parselist('O', ['&' + pname], [pname])
+        info.arglist.append(pname)
+    def write_return(self, ptype, ownsreturn, info):
+        info.varlist.add("PyObject", "*ret")
+        if ownsreturn:
+            info.codeafter.append('    if (ret) {\n'
+                                  '       return ret;\n'
+                                  '    }\n'
+                                  '    Py_INCREF(Py_None);\n'
+                                  '    return Py_None;')
+        else:
+            info.codeafter.append('    if (!ret) ret = Py_None;\n'
+                                  '    Py_INCREF(ret);\n'
+                                  '    return ret;')
+
+class CairoArg(ArgType):
+    def write_param(self, ptype, pname, pdflt, pnull, info):
+        info.varlist.add('PycairoContext', '*' + pname)
+        info.add_parselist('O!', ['&PycairoContext_Type', '&' + pname], [pname])
+        info.arglist.append('%s->ctx' % pname)
+    def write_return(self, ptype, ownsreturn, info):
+        info.varlist.add("cairo_t", "*ret")
+        if ownsreturn:
+            info.codeafter.append('    return PycairoContext_FromContext(ret, NULL, NULL);')
+        else:
+            info.codeafter.append('    cairo_reference(ret);\n'
+                                  '    return PycairoContext_FromContext(ret, NULL, NULL);')
+
+
 class ArgMatcher:
     def __init__(self):
 	self.argtypes = {}
+	self.reverse_argtypes = {}
+	self.reverse_rettypes = {}
 
     def register(self, ptype, handler):
 	self.argtypes[ptype] = handler
+    def register_reverse(self, ptype, handler):
+	self.reverse_argtypes[ptype] = handler
+    def register_reverse_ret(self, ptype, handler):
+	self.reverse_rettypes[ptype] = handler
+        
     def register_enum(self, ptype, typecode):
         if typecode is None:
             typecode = "G_TYPE_NONE"
@@ -686,6 +827,39 @@ class ArgMatcher:
             if ptype[:8] == 'GdkEvent' and ptype[-1] == '*':
                 return self.argtypes['GdkEvent*']
             raise
+    def _get_reverse_common(self, ptype, registry):
+        props = dict(c_type=ptype)
+        try:
+            return registry[ptype], props
+        except KeyError:
+            try:
+                handler = self.argtypes[ptype]
+            except KeyError:
+                if ptype.startswith('GdkEvent') and ptype.endswith('*'):
+                    handler = self.argtypes['GdkEvent*']
+                else:
+                    raise
+            if isinstance(handler, ObjectArg):
+                return registry['GObject*'], props
+            elif isinstance(handler, EnumArg):
+                props['typecode'] = handler.typecode
+                props['enumname'] = handler.enumname
+                return registry['GEnum'], props
+            elif isinstance(handler, FlagsArg):
+                props['typecode'] = handler.typecode
+                props['flagname'] = handler.flagname
+                return registry['GFlags'], props
+            elif isinstance(handler, BoxedArg):
+                props['typecode'] = handler.typecode
+                props['typename'] = handler.typename
+                return registry['GBoxed'], props
+            else:
+                raise
+    def get_reverse(self, ptype):
+        return self._get_reverse_common(ptype, self.reverse_argtypes)
+    def get_reverse_ret(self, ptype):
+        return self._get_reverse_common(ptype, self.reverse_rettypes)
+
     def object_is_a(self, otype, parent):
         if otype == None: return 0
         if otype == parent: return 1
@@ -702,7 +876,9 @@ arg = StringArg()
 matcher.register('char*', arg)
 matcher.register('gchar*', arg)
 matcher.register('const-char*', arg)
+matcher.register('char-const*', arg)
 matcher.register('const-gchar*', arg)
+matcher.register('gchar-const*', arg)
 matcher.register('string', arg)
 matcher.register('static_string', arg)
 
@@ -722,27 +898,32 @@ matcher.register('gunichar', arg)
 arg = IntArg()
 matcher.register('int', arg)
 matcher.register('gint', arg)
-matcher.register('guint', arg)
 matcher.register('short', arg)
 matcher.register('gshort', arg)
 matcher.register('gushort', arg)
-matcher.register('long', arg)
-matcher.register('glong', arg)
-matcher.register('gboolean', arg)
-
+matcher.register('gsize', SizeArg())
+matcher.register('gssize', SSizeArg())
 matcher.register('guint8', arg)
 matcher.register('gint8', arg)
 matcher.register('guint16', arg)
 matcher.register('gint16', arg)
 matcher.register('gint32', arg)
+matcher.register('GTime', arg)
 
-# If the system maxint is smaller than unsigned int, we need to use
-# Long objects with PyLong_AsUnsignedLong
-if sys.maxint >= (1L << 32):
-    matcher.register('guint32', arg)
-else:
-    arg = ULongArg()
-    matcher.register('guint32', arg)
+arg = LongArg()
+matcher.register('long', arg)
+matcher.register('glong', arg)
+
+arg = UIntArg()
+matcher.register('guint', arg)
+
+arg = BoolArg()
+matcher.register('gboolean', arg)
+
+arg = TimeTArg()
+matcher.register('time_t', arg)
+
+matcher.register('guint32', UInt32Arg())
 
 arg = ULongArg()
 matcher.register('gulong', arg)
@@ -776,9 +957,12 @@ matcher.register('GtkTreePath*', GtkTreePathArg())
 matcher.register('GdkRectangle*', GdkRectanglePtrArg())
 matcher.register('GtkAllocation*', GdkRectanglePtrArg())
 matcher.register('GdkRectangle', GdkRectangleArg())
+matcher.register('PyObject*', PyObjectArg())
 
 matcher.register('GdkNativeWindow', ULongArg())
 
 matcher.register_object('GObject', None, 'G_TYPE_OBJECT')
 
 del arg
+
+matcher.register('cairo_t*', CairoArg())
