@@ -701,9 +701,15 @@ create_property (const gchar  *prop_name,
     case G_TYPE_ENUM:
 	{
 	    gint default_value;
-
-	    if (!PyArg_ParseTuple(args, "i", &default_value))
+	    PyObject *pydefault;
+	    
+	    if (!PyArg_ParseTuple(args, "O", &pydefault))
 		return NULL;
+	    
+	    if (pyg_enum_get_value(prop_type, pydefault,
+				   (gint *)&default_value))
+		return NULL;
+	    
 	    pspec = g_param_spec_enum (prop_name, nick, blurb,
 				       prop_type, default_value, flags);
 	}
@@ -711,9 +717,15 @@ create_property (const gchar  *prop_name,
     case G_TYPE_FLAGS:
 	{
 	    guint default_value;
+	    PyObject *pydefault;
 
-	    if (!PyArg_ParseTuple(args, "i", &default_value))
+	    if (!PyArg_ParseTuple(args, "O", &pydefault))
 		return NULL;
+	    
+	    if (pyg_flags_get_value(prop_type, pydefault,
+				    (gint *)&default_value))
+		return NULL;
+	    
 	    pspec = g_param_spec_flags (prop_name, nick, blurb,
 					prop_type, default_value, flags);
 	}
@@ -962,11 +974,6 @@ static PyObject *
 _wrap_pyg_type_register(PyObject *self, PyObject *args)
 {
     PyTypeObject *class;
-    PyObject *gtype;
-
-    if (PyErr_Warn(PyExc_DeprecationWarning, "gobject.type_register is deprecated;"
-                   " now types are automatically registered"))
-        return NULL;
 
     if (!PyArg_ParseTuple(args, "O:gobject.type_register", &class))
 	return NULL;
@@ -976,16 +983,25 @@ _wrap_pyg_type_register(PyObject *self, PyObject *args)
     }
 
       /* Check if type already registered */
-    gtype = PyObject_GetAttrString((PyObject *)class, "__gtype__");
-    if (gtype == NULL) {
-        PyErr_Clear();
-          /* not registered */
+    if (pyg_type_from_object((PyObject *) class) ==
+        pyg_type_from_object((PyObject *) class->tp_base))
+    {
         if (pyg_type_register(class, NULL))
             return NULL;
-    } else
+    } else {
+        gchar *msg;
           /* already registered */
-        Py_DECREF(gtype);
-
+        msg = g_strdup_printf("Class %s is already GObject-registered; "
+			      "Please note that classes containing any of the "
+			      "attributes __gtype_name__, __gproperties__, or "
+			      "__gsignals__ are now automatically registered.",
+			      class->tp_name);
+        if (PyErr_Warn(PyExc_DeprecationWarning, msg)) {
+            g_free(msg);
+            return NULL;
+        }
+        g_free(msg);
+    }
     Py_INCREF(class);
     return (PyObject *) class;
 }
@@ -1146,10 +1162,7 @@ pyg_type_register(PyTypeObject *class, char *type_name)
       /* propagate new constructor API compatility flag from parent to child type */
     has_new_constructor_api = g_type_get_qdata(parent_type,
                                                pygobject_has_updated_constructor_key);
-    if (has_new_constructor_api == NULL)
-        g_warning("Constructor wrapper for %s needs to be updated to the new API",
-                  g_type_name(parent_type));
-    else
+    if (has_new_constructor_api != NULL)
         g_type_set_qdata(instance_type, pygobject_has_updated_constructor_key,
                          has_new_constructor_api);
 
@@ -1637,10 +1650,11 @@ pyg_object_new (PyGObject *self, PyObject *args, PyObject *kwargs)
     g_type_class_unref(class);
     
     if (obj)
-	self = (PyGObject *) pygobject_new ((GObject *)obj);
+	self = (PyGObject *) pygobject_new_full((GObject *)obj, FALSE);
     else
         self = NULL;
     g_object_unref(obj);
+    pygobject_sink(obj);
 
     return (PyObject *) self;
 }
@@ -1882,18 +1896,9 @@ pyg_source_remove(PyObject *self, PyObject *args)
 }
 
 static PyObject *
-pyg_main_context_default (PyObject *unused)
+pyg_main_context_default(PyObject *unused)
 {
-    PyGMainContext *self;
-
-    self = (PyGMainContext *)PyObject_NEW(PyGMainContext,
-					  &PyGMainContext_Type);
-    if (self == NULL)
-	return NULL;
-
-    self->context = g_main_context_default();
-    return (PyObject *)self;
-
+    return pyg_main_context_new(g_main_context_default());
 }
 
 static int pyg_thread_state_tls_key = -1;
@@ -1914,12 +1919,6 @@ pyg_enable_threads ()
 	pygobject_api_functions.threads_enabled = TRUE;
 	pyg_thread_state_tls_key = PyThread_create_key();
     }
-    if (PYGIL_API_IS_BUGGY && !use_gil_state_api) {
-	PyThreadState* state;
-	state = PyThreadState_Get();
-	if ( state != NULL )
-	    PyThread_set_key_value(pyg_thread_state_tls_key, state);
-    }
 
     return 0;
 #else
@@ -1929,54 +1928,16 @@ pyg_enable_threads ()
 #endif
 }
 
-static PyThreadState *
-pyg_find_thread_state (void)
-{
-    PyThreadState* state;
-    
-    if (pyg_thread_state_tls_key == -1)
-	return NULL;
-    state = PyThread_get_key_value(pyg_thread_state_tls_key);
-    if (state == NULL) {
-	state = PyGILState_GetThisThreadState();
-	if (state != NULL)
-	    PyThread_set_key_value(pyg_thread_state_tls_key, state);
-    }
-    return state;
-}
-
 static int
 pyg_gil_state_ensure_py23 (void)
 {
-    if (PYGIL_API_IS_BUGGY && !use_gil_state_api) {
-	PyThreadState* state = pyg_find_thread_state();
-
-	if (state == NULL)
-	    return PyGILState_LOCKED;
-    
-	if (state == _PyThreadState_Current)
-	    return PyGILState_LOCKED;
-	else {
-	    PyEval_RestoreThread(state);
-	    return PyGILState_UNLOCKED;
-	}
-    } else {
-	return PyGILState_Ensure();
-    }
+    return PyGILState_Ensure();
 }
 
 static void
 pyg_gil_state_release_py23 (int flag)
 {
-    if (PYGIL_API_IS_BUGGY && !use_gil_state_api) {
-	if (flag == PyGILState_UNLOCKED) {
-	    PyThreadState* state = pyg_find_thread_state();
-	    if (state != NULL)
-		PyEval_ReleaseThread(state);
-	}
-    } else {
-	PyGILState_Release(flag);
-    }
+    PyGILState_Release(flag);
 }
 
 static PyObject *
@@ -2748,9 +2709,10 @@ initgobject(void)
 
     REGISTER_TYPE(d, PyGTypeWrapper_Type, "GType");
 
-    PY_TYPE_OBJECT = g_boxed_type_register_static("PyObject",
-						  pyobject_copy,
-						  pyobject_free);
+    if (!PY_TYPE_OBJECT)
+	PY_TYPE_OBJECT = g_boxed_type_register_static("PyObject",
+						      pyobject_copy,
+						      pyobject_free);
 
     gerror_exc = PyErr_NewException("gobject.GError", PyExc_RuntimeError,NULL);
     PyDict_SetItemString(d, "GError", gerror_exc);
